import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt

#TODO: replace all paths to input data (load love numbers, ddk binaries) with variable to set at the top

def readicgem(ifile: str) -> tuple[np.ndarray, np.ndarray, dict, np.ndarray, np.ndarray]:
    '''
    READICGEM: Reads potential coefficients in ICGEM-format from ASCII.

    IN:
        - ifile: path/name of input-file

    OUT:
        - scs   : coefficients in cs-format (signal)
        - ncs   : coeff. errors in cs-format (signal)
        - header: dict. containing header entries
        - scst  : dot-coefficients in cs-format
        - ncst  : dot-errors in cs-format
    
    USAGE EXAMPLE:
        scs, ncs, header, scst, ncst = readicgem('/home/foo/bar.gfc')

    HISTORY:
        - 10/10/2005 Martin Wermuth, IAPG, TU-Munich
        - 19/01/2010 Michael Murboeck, IAPG, TU-Munich
            separation of formal and calibrated errors
            dot-coefficients
            cs_format instead of ma2cs
        - 09/07/2024 Linus Shihora, GFZ, Potsdam
            porting to python
    '''


    #Read header
    header = dict()

    f = open(ifile, 'r')

    #Skip lines until header begins
    for index, line in enumerate(f):
        if 'begin_of_head' in line:
            break
    
    #Start reading header entries until end_of_head
    for index, line in enumerate(f):
        if 'end_of_head' in line:
            break
        line = line.split()
        
        #Skip empty lines
        if len(line) == 0:
            continue
        elif len(line) == 2:
            header[line[0]] = line[1]
        else:
            header[line[0]] = line[1:]

    

    if header['errors'] == 'no' or header['errors'] == 'not included':
        #Initialise coefficient arrays
        lmax = int(header['max_degree'])
        
        shape = (lmax+1, lmax+1)
        c_lm = np.zeros(shape)
        s_lm = np.zeros(shape)

        c_lm_t = np.zeros(shape)
        s_lm_t = np.zeros(shape)
        lmax_t = 0

        #Continue loop through ifile after header
        for index, line in enumerate(f):
            line = line.split()

            if line[0] == 'dot':
                l = int(line[1])
                m = int(line[2])
                c_lm_t[l,m] = float(line[3])
                s_lm_t[l,m] = float(line[4])
                lmax_t = max([lmax_t, l])
            
            elif line[0] == 'gfc' or line[0] == 'gfct':
                l = int(line[1])
                m = int(line[2])
                c_lm[l,m] = float(line[3])
                s_lm[l,m] = float(line[4])
            
            if line[0] == 'gfct': header['reference_epoch'] = line[-1]

        c_lm_t = c_lm_t[:lmax_t+1, :lmax_t+1]
        s_lm_t = s_lm_t[:lmax_t+1, :lmax_t+1]

        #Transform to cs-format
        scs = cs_format(np.stack((c_lm, s_lm)), 1)
        scs_t = cs_format(np.stack((c_lm_t, s_lm_t)), 1)
        ncs = []
        ncs_t = []



    else: #Consider errors
        #Initialise coefficient arrays
        lmax = int(header['max_degree'])
        
        shape = (lmax+1, lmax+1)
        c_lm = np.zeros(shape)
        s_lm = np.zeros(shape)
        sigma_c_lm = np.zeros(shape)
        sigma_s_lm = np.zeros(shape)

        c_lm_t = np.zeros(shape)
        s_lm_t = np.zeros(shape)
        sigma_c_lm_t = np.zeros(shape)
        sigma_s_lm_t = np.zeros(shape)
        lmax_t = 0

        #Continue loop through ifile after header
        for index, line in enumerate(f):
            line = line.split()
            
            if line[0] == 'dot':
                l = int(line[1])
                m = int(line[2])
                c_lm_t[l,m] = float(line[3])
                s_lm_t[l,m] = float(line[4])

                if header['errors'] == 'calibrated_and_formal':
                    sigma_c_lm_t[l,m] = float(line[7])
                    sigma_s_lm_t[l,m] = float(line[8])
                    header['errors'] = 'formal'
                else:
                    sigma_c_lm_t[l,m] = float(line[5])
                    sigma_s_lm_t[l,m] = float(line[6])
                lmax_t = max([lmax_t, l])
            
            elif line[0] == 'gfc' or line[0] == 'gfct':
                l = int(line[1])
                m = int(line[2])
                c_lm[l,m] = float(line[3])
                s_lm[l,m] = float(line[4])

                if header['errors'] == 'calibrated_and_formal':
                    sigma_c_lm[l,m] = float(line[7])
                    sigma_s_lm[l,m] = float(line[8])
                else:
                    sigma_c_lm[l,m] = float(line[5])
                    sigma_s_lm[l,m] = float(line[6])

            if line[0] == 'gfct': header['reference_epoch'] = line[-1]
    
        c_lm_t = c_lm_t[:lmax_t+1, :lmax_t+1]
        s_lm_t = s_lm_t[:lmax_t+1, :lmax_t+1]
        sigma_c_lm_t = sigma_c_lm_t[:lmax_t+1, :lmax_t+1]
        sigma_s_lm_t = sigma_s_lm_t[:lmax_t+1, :lmax_t+1]

        #Transform to cs-format
        scs = cs_format(np.stack((c_lm, s_lm)), 1)
        scs_t = cs_format(np.stack((c_lm_t, s_lm_t)), 1)
        ncs = cs_format(np.stack((sigma_c_lm, sigma_s_lm)), 1)
        ncs_t = cs_format(np.stack((sigma_c_lm_t, sigma_s_lm_t)), 1)

    f.close()

    return scs, ncs, header, scs_t, ncs_t







def cs_format(coeff: list | np.ndarray, num_output: int) -> np.ndarray | tuple[np.ndarray, np.ndarray]:
    '''
    CS_FORMAT: Function to transform format of spherical harmonic coefficients

    IN:
        - coeff     : coefficients of spherical harmonics either [cs] or [c,s]
        - num_output: number of outputs

    OUT:
        - cs, s: reordered coefficients

    FORMATS:
        - |C\\S|     : [n,n] matrix with c_nm in the lower triangle and s_nm in the upper 
                      triangle (c_n0 in first column, s_n1 in first row)
        - |/SC\\|    : [n,2n] or [n,2n-1] matrix with s_nm in lower right triangle on the left half 
                      and c_nm in lower left triangle of the right half (s_n0 as zeros in column left 
                      of center or missing)
        - |C\\| |S\\| : two [n,n] matrices with c_nm and s_nm each in the lower triangular matrix (s_nm
                      as [n,n-1] matrix if s_n0 are missing, otherwise zeros)
    
    TRANSORMATIONS:
        - 1 input, 1 output     : |C\\S|            --> |/S C\\|  [n,2n-1] matrix
                                  |/SC\\|           --> |C\\S|
        - 1 input, 2 outputs    : |C\\S|, |/SC\\|   --> |C\\| |S\\| two [n,n] matrices
        - 2 inputs, 1 oputput   : |C\\||S\\|        --> |C\\S|

    USAGE EXAMPLES:
        sc      = cs_format(cs,1)
        cs      = cs_format(sc,1)
        c, s    = cs_format(cs,2)
        c, s    = cs_format(sc,2)
        cs      = cs_format(np.array([c,s]),1)

    HISTORY:
        - 12/2008 Michael Murboeck
        - 09/07/2024 Linus Shihora, GFZ, Potsdam
            porting to python
    TODO:
        -
    '''

    coeff = np.asarray(coeff)

    if coeff.ndim > 3:
        raise ValueError('Input array must be either |C\\S|, |/SC\\| or [|C\\|,|S\\|].')

    if not isinstance(num_output, int):
        raise TypeError('num_outputs must be an integer.')

    if num_output not in [1,2]:
        raise ValueError('Number of outputs must be either 1 or 2.')


    if coeff.ndim == 3:
        c = coeff[0]
        s = coeff[1]

        c_rows, c_cols = c.shape
        s_rows, s_cols = s.shape

        #Ensure array sizes are usable
        if c_rows != c_cols:
            raise ValueError('Number of rows and columns of [c] must match.')

        if s_rows != c_rows:
            raise ValueError('Number of rows in [c] and [s] must match.')
        
        n_max = c_rows - 1

        if s_cols == (s_rows - 1): 
            s = np.hstack(np.zeros((n_max+1,1)), s)
        elif s_cols == s_rows:
            pass
        else:
            raise Exception('Columns of s must match rows or (rows-1).')
        
        if num_output == 1:
            s_sub = s[:,1:]
            s_sub = s_sub.T
            zero_row = np.zeros((1, n_max + 1))
            c = c + np.vstack((s_sub, zero_row))
        else:
            raise Exception('Unknown number of outputs. For [c,s]-input, num_output must be 1.')
        
        return c

    else: #coeff.ndim == 2
        rows, cols = coeff.shape

        if cols == rows:
            n_max = rows - 1
            c = np.tril(coeff) #creates lower triangular array
            s_t = np.triu(coeff, 1).T #creates upper triangular array skipping diagonal
            s = np.hstack((np.zeros((n_max+1,1)), s_t[:,:-1]))

            if num_output == 1:
                s_flipped = np.fliplr(s[:,1:])
                c = np.hstack((s_flipped, c))
                return c

        elif cols == 2*rows:
            n_max = rows - 1
            c = coeff[:, (n_max+1):]
            s = np.fliplr(coeff[:,n_max+1])
            
            if num_output == 1:
                for ro in range(n_max):
                    c[ro, ro+1:] = s[ro+1:, ro+1].T
                    return c
        
        elif cols == 2*rows - 1:
            n_max = rows - 1
            c = coeff[:, n_max:]
            s_sub = np.fliplr(coeff[:,:n_max])
            zero_col = np.zeros((n_max+1, 1))
            s = np.hstack((zero_col, s_sub))

            if num_output == 1:
                s_sub = s[:,1:]
                s_sub = s_sub.T
                zero_row = np.zeros((1, n_max + 1))
                c = c + np.vstack((s_sub, zero_row))
                return c
        
        else:
            raise Exception('Columns of cs must match rows or 2*rows or 2*rows-1.')
        
        return c, s




def legnorm(l: np.ndarray, m: int, colat: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    '''
    LEGNORM: Computes normalized associated Legendre polynomials and their derivatives.

    IN:
        - l     : Array of degrees of polynomials.
        - m     : Order of polynomials.
        - colat : Vector of colatitudes in degrees.

    OUT:
        - p : Array of normalised associated Legendre polynomials
        - dp: Array of the derivatives of the normalised associated Legendre polynomials

    USAGE EXAMPLE:
        p, dp = legnorm(l, m, col)

    HISTORY:
        - 31/10/2011, Michael Murboeck, Muenchen
        - 11/07/2024, Linus Shihora, GFZ, Potsdam
            porting to python
    TODO:
        - Does not work for len(l) < 3
        - Make sure that single l inputs are possible
        - Input must be numpy array. Make sure it works for lists too
    '''
    
    if m < 0:
        raise ValueError('Order m must be > 0.')

    if l.ndim != 1: 
        raise ValueError('Array of degrees l must be of dimension 1.')
    
    if not np.all(l.astype(int) == l):
        raise TypeError('Array of degrees l must contain only integers.')
    
    if not np.any(l >= m): 
        raise ValueError('All degrees l must be larger than the order m.')
    
    colat = np.asarray(colat) * np.pi / 180
    if colat.ndim == 1:
        colat = colat[:, np.newaxis]
    
    lmax = np.max(l)
    t = np.cos(colat)
    y = np.sqrt(1 - t**2)
    p00 = np.ones(t.shape)
    dp00 = np.zeros(t.shape)
    p = np.zeros((len(t), lmax - m + 1))
    dp = np.zeros((len(t), lmax - m + 1))

    def wlmfun(l: int, m: int) -> int:
        '''
        Computes normalizing coefficients for Legendre functions.
        '''
        if l == m:
            if l == 0:
                return 1
            elif l == 1:
                return np.sqrt(3)
            else:
                return np.sqrt((2 * l + 1) / (2 * l))
        else:
            return np.sqrt((2 * l + 1) / (l + m) * (2 * l - 1) / (l - m))
    
    if m == 0:
        plm3 = np.sqrt(3) * t  # P10
        plm2 = np.sqrt(5) * (3 * t**2 - 1) / 2  # P20
        p[:, :3] = np.hstack([p00, plm3, plm2])
        dplm3 = -y * wlmfun(1, 0)
        dplm2 = wlmfun(2, 0) * (-y * plm3 + t * dplm3)
        dp[:, :3] = np.hstack([dp00, dplm3, dplm2])
    elif m == 1:
        plm3 = np.sqrt(3) * y  # P11
        plm2 = np.sqrt(15) * t * y  # P21
        p[:, :2] = np.hstack([plm3, plm2])
        dplm3 = wlmfun(1, 1) * t
        dplm2 = wlmfun(2, 1) * (-y * plm3 + t * dplm3)
        dp[:, :2] = np.hstack([dplm3, dplm2])
    else:
        plm3 = np.zeros(t.shape)  # Pm-1,m
        plm2 = np.sqrt(15) * y**2 / 2  # P22
        dplm3 = np.zeros(t.shape)
        dplm2 = wlmfun(2, 2) * (np.sqrt(3) * t * y + wlmfun(1, 1) * y * t)
        for n in range(4, m + 1):
            dplm2 = wlmfun(n - 1, n - 1) * (t * plm2 + y * dplm2)
            plm2 = np.sqrt((2 * n - 1) / (2 * n - 2)) * y * plm2
        
        p[:, 0] = plm2[:, 0]
        dp[:, 0] = dplm2[:, 0]

    if lmax > max(2, m):
        for n in range(max(2, m) + 2, lmax + 2):
            plm1 = np.sqrt((2 * n - 1) / (n - m - 1) * (2 * n - 3) / (n + m - 1)) * t * plm2 - \
                   np.sqrt((2 * n - 1) / (2 * n - 5) * (n + m - 2) / (n + m - 1) * (n - m - 2) / (n - m - 1)) * plm3
            dplm1 = wlmfun(n - 1, m) * (-y * plm2 + t * dplm2 - dplm3 / wlmfun(n - 2, m))
            p[:, n - m - 1] = plm1[:, 0]
            dp[:, n - m - 1] = dplm1[:, 0]
            plm3, plm2 = plm2, plm1
            dplm3, dplm2 = dplm2, dplm1

    p = p[:, l - m]
    dp = dp[:, l - m]

    return p, dp






def shs(gco: np.ndarray, 
        func: list | np.ndarray, 
        colat: np.ndarray = None, 
        lon: np.ndarray = None, 
        AE: int = 6378137, 
        ALT: int = 0, 
        **kwargs) -> np.ndarray:
    '''
    SHS: Function to synthesise spherical harmonics onto a global grid

    IN:
        - gco       : Gravity coefficients in cs-format, i.e. a [l_max+1,l_max+1] matrix with 
                        c_lm in the lower triangular and s_lm in the upper triangular
        - func      : Gravity functionals. Vector with up to 5 entries. Default: func=[1]. Possible entries:
            - 1     : Geoid undulation in m.
            - 2     : Gravity anomaly in mGal.
            - 3     : Vertical gravity gradient in E.
            - 4     : Total water storage in mm EWH.
            - 5     : No dimension.
            - 6     : Gravity disturbance in mGal.
            - 7     : Pressure in Pa.
            - 8     : Vertical crustal deformation in m.
        - colat     : Co-latitude vector in degrees. Default: 0.5:179.5
        - lon       : Longitude vector in degrees. Default: 0.5:359.5
        - AE        : Earth radius in m. Default: ae=6378137
        - ALT       : Satellite altitude. Default: alt=0
        - **kwargs  : Optional key-word arguments that have default-values. Possible argumets:
            - GM    : Earth gravity constant multiplied by Earths mass in m^3/s^2. Default: GM=3.986005e14
            - l_max : Max. SH. degree. Default: (gco.shape[0] - 1)

    OUT:
        - global_grd: Output array containing gridded fields. Array of shape [len(colat), len(lat), len(func)]

    USAGE EXAMPLE:
        grav_grid = shs(gco, [1])
        grav_grid = shs(gco, [1],  colat=colat, lon=lon, func=[1,4,7], l_max=180)

    HISTORY:
        - Original author unknown
        - 07/2024 Simon Schiller, TUM, Munich
            porting to python
        - 11/07/2024 Linus Shihora, GFZ, Potsdam
            formatting & checks
    '''

    #Check for plausible input
    if not isinstance(func, list | np.ndarray):
        raise TypeError('Gravity functional func must be a list or np.ndarray.')
    elif any(f not in range(1, 9) for f in func):
        raise ValueError('all entries in func must be in 1:8')
    

    row, col = gco.shape
    if row != col: 
        raise ValueError('Input gravity coefficients must be a square matrix.')


    if not isinstance(colat, list | np.ndarray):
        if colat == None:
            colat = np.arange(0.5, 180)
        else:
            raise TypeError('Colatitude must be a list or np.ndarray.')


    if not isinstance(lon, list | np.ndarray):
        if lon == None:
            lon = np.arange(0.5, 360)
        else:
            raise TypeError('Longitude must be a list or np.ndarray.')


    #Check for unexpected kwargs
    valid_args = {'GM', 'l_max'}
    undefined_args = [key for key, _ in kwargs.items() if key not in valid_args]
    if undefined_args:
        raise TypeError(f'Unexpected kwargs: {undefined_args}')


    #Get possible kwargs; otherwise set default
    GM = kwargs.get('GM', 3.986005e14)
    l_max = kwargs.get('l_max', (gco.shape[0] - 1))
    
    if not isinstance(l_max, int):
        raise TypeError('l_max must be an integer.')
    
    if l_max > (row - 1):
        raise ValueError('l_max larger than gravity coefficient matrix.')

    if l_max <= 0:
        raise ValueError('l_max must be > 0.')



    #Set constants
    G = 6.67259e-11


    #Create all matrices required for the calculation (d/o, sin & cos coefficients):
    l_vec = np.arange(0, l_max + 1)
    m_vec = np.arange(0, l_max + 1).reshape(-1, 1)
    
    cos_coeff = np.tril(gco)
    sin_coeff = np.triu(gco, 1).T
    sin_coeff = np.hstack((sin_coeff[:, -1:], sin_coeff[:, :-1]))

    cos_mlon = np.cos(m_vec * lon * np.pi / 180)
    sin_mlon = np.sin(m_vec * lon * np.pi / 180)

    global_grd = np.zeros((len(colat), len(lon), len(func)))


    #Loop over all functionals
    for i in range(len(func)):
        Am = np.zeros((len(colat), l_max+1))
        Bm = np.zeros((len(colat), l_max+1))

        for k in range(len(m_vec)):
            p_func, _  = legnorm(l_vec[k:], m_vec[k, 0], colat)
            attfact = (AE / (AE + ALT)) ** (l_vec[k:] + 1)

            if func[i] == 1: #Geoid undulation in m
                c_func = AE
                l_func = np.ones(len(l_vec[k:]))

            elif func[i] == 2: #Gravity anomaly in mGal.
                c_func = 1e5 * GM / AE**2
                l_func = l_vec[k:] - 1

            elif func[i] == 3: #Vertical gravity gradient in E.
                c_func = 1e9 * GM / AE**3
                l_func = (l_vec[k:] + 1) * (l_vec[k:] + 2)

            elif func[i] == 4: #Total water storage in mm EWH.
                lln = np.loadtxt('./data/loadlove_farrell.dat')
                ks = np.append([0], lln[:l_max, 2].T)
                c_func = GM / G / 4 / np.pi / AE**2
                l_func = (2*l_vec[k:] + 1) / (ks[k:] + 1)

            elif func[i] == 5: #No dimension.
                c_func = 1
                l_func = np.ones(len(l_vec[k:]))

            elif func[i] == 6: #Gravity disturbance in mGal.
                c_func = 1e5 * GM / AE**2
                l_func = l_vec[k:] + 1

            elif func[i] == 7: #Pressure in Pa.
                lln = np.loadtxt('./data/loadlove_farrell.dat')
                ks = np.append([0], lln[:l_max, 2].T)
                c_func = 9.81 * GM / G / 4 / np.pi / AE**2
                l_func = (2*l_vec[k:] + 1) / (ks[k:] + 1)

            elif func[i] == 8: #Vertical crustal deformation in m.
                lln = np.loadtxt('./data/loadlove_farrell.dat')
                hs = np.append([0,0], lln[1:l_max, 1].T)
                ks = np.append([0], lln[:l_max, 2].T)
                c_func = AE
                l_func = hs[k:] / (ks[k:] + 1)

            #Calculate final grid
            l_func = l_func * attfact
            tmp = p_func * np.tile(l_func, (len(colat),1))
            Am[:, k] = tmp @ cos_coeff[k:l_max + 1, k]

            if m_vec[k] != 0:
                Bm[:, k] = tmp @ sin_coeff[k:l_max + 1, k]

        global_grd[:,:,i] = c_func * (Am @ cos_mlon + Bm @ sin_mlon)
    
    return global_grd

    






def reduce_normpot_grs80(scs: np.ndarray | list, GM: float = 3.986005e14, AE: int = 6378137) -> np.ndarray:
    '''
    REDUCE_NORMPOT_GRS80: Function to subtract / reduce the notmal potential from a given gravity field model

    IN:
        - scs       : Gravity coefficients in cs-format, i.e. a [l_max+1,l_max+1] matrix with 
                        c_lm in the lower triangular and s_lm in the upper triangular
        - Optionals:
            - GM    : Earth gravity constant multiplied by Earths mass in m^3/s^2. Default: GM=3.986005e14
            - ae    : Earth radius in m. Default: ae=6378137

    OUT:
        - scs_red   : Disturbin potential (w.r.t. GRS80) in cs-format

    USAGE EXAMPLE:
        dist_pot = reduce_normpot_grs80(scs, gm = 3.986005e14, ae = 6378137)
        dist_pot = reduce_normpot_grs80(scs)

    HISTORY:
        - Original author unknown
        - 07/2024 Simon Schiller, TUM, Munich
            porting to python
        - 13/08/2024 Linus Shihora, GFZ, Potsdam
            formatting
    '''

    #Check input types
    if not isinstance(GM, float):
        raise TypeError('GM must be a float.')
    if not isinstance(AE, int):
        raise TypeError('AE must be an integer.')

    #Check whether the dimensions of the scs model are identical (nxn matrix)
    if scs.shape[0] != scs.shape[1]:
        raise ValueError('Unknown shape. Gravity field input must be provided in cs-format.')

    #Calculation of the maximum degree/order and the constants for the GRS80 gravity field
    n_max = scs.shape[0] - 1
    A_GRS80, GM_GRS80, _, _, scs_GRS80 = const_grs80(n_max)

    #Adaptation of the GRS80 gravity field to GM and ae and reduction from the given gravity field
    for n in range(9):
        scs_GRS80[n,0] = scs_GRS80[n,0] * (GM / GM_GRS80) * (AE / A_GRS80) ** n
    
    return scs - scs_GRS80



def const_grs80(n_max: int) -> tuple[int, float, float, list, np.ndarray]:
    '''
    CONST_GRS80: Define and derive constants of GRS80 reference ellipsoid.

    IN:
        - n_max     : maximum degree

    OUT:
        - A_GRS80, GM_GRS80, W_GRS80, J_GRS80   : Constants of normal potential
        - scs_GRS80                             : Normal potential

    USAGE EXAMPLE:
        A_GRS80, GM_GRS80, W_GRS80, J_GRS80, scs_GRS80 = const_grs80(n_max)

    HISTORY:
        - Original author unknown
        - 07/2024 Simon Schiller, TUM, Munich
            porting to python
        - 13/08/2024 Linus Shihora, GFZ, Potsdam
            formatting
    '''

    #Definition of the constants of the GRS80 normal potential
    A_GRS80 = 6378137
    W_GRS80 = 7292115e-11
    GM_GRS80 = 3986005e8

    J2_GRS80 = 108263e-8
    J4_GRS80 = -0.00000237091222
    J6_GRS80 = 0.00000000608347
    J8_GRS80 = -0.00000000001427
    J_GRS80 = [J2_GRS80, J4_GRS80, J6_GRS80, J8_GRS80]


    #Determination of the sine and cosine coefficients of the GRS80 normal potential from the values of J:
    scs_GRS80 = np.zeros((n_max+1, n_max+1))

    for i in range(2,9,2):
        if i > n_max:
            break
        scs_GRS80[i,0] = (-1) * J_GRS80[(i // 2) - 1] / np.sqrt(2 * i + 1)
    
    scs_GRS80[0,0] = 1

    return A_GRS80, GM_GRS80, W_GRS80, J_GRS80, scs_GRS80





def gaussian(n_max: int, radius_filter: int) -> np.ndarray:
    '''
    GAUSSIAN: Compute weights for Gaussian filtering for every degree up to n_max.
              The Gaussian weighting function is calculated based on eq. (4) of
              'Chambers et al., JGR, 2006, Observing seasonal steric sea level 
              variations with GRACE and satellite altimetry'

    IN:
        - n_max         : maximum degree
        - radius_filter : radius of gaussian smoothing in km, 
                          i.e. half-width of Gaussian weighting function

    OUT:
        - weights       : weighting coefficients as np.ndarray of shape (n_max+1, 1)

    USAGE EXAMPLE:
        W = gaussian(180, 500)

    HISTORY:
        - 18/12/2015 Feng Wei, State Key Laboratory of Geodesy and Earth's Dynamics &
            Institute of Geodesy and Geophysics, Chinese Academy of Sciences
        - 15/08/2024 Linus Shihora, GFZ, Potsdam
            porting to python
    '''

    #Check input types
    if not isinstance(n_max, int):
        raise TypeError('n_max must be an integer.')
    if not isinstance(radius_filter, int):
        raise TypeError('Filter radius must be an integer.')
    
    #Check for plausible inputs
    if n_max < 2:
        raise ValueError('n_max must be higher than 2.')
    if radius_filter < 0:
        raise ValueError('Filter radius can not be negative.')
    

    #Initialise arrays
    weights = np.zeros(n_max + 1)

    for i in range(n_max + 1):
        weights[i] = np.exp(-(i*radius_filter/6371)**2 / (4*np.log(2)))

    return weights




def filter_sh_gaussian(scs: np.ndarray | list, radius_filter: int) -> np.ndarray:
    '''
    FILTER_SH_GAUSSIAN: Apply a Gaussian smoothing to spherical harmonic coefficients.
                        Requires function 'gaussian' to calculate weights.

    IN:
        - scs           : Gravity coefficients in cs-format, i.e. a [l_max+1,l_max+1] matrix with 
                          c_lm in the lower triangular and s_lm in the upper triangular
        - radius_filter : radius of gaussian smoothing in km, 
                          i.e. half-width of Gaussian weighting function

    OUT:
        - scs_filter    : Gravity coefficients in cs-format, 
                          i.e. a [l_max + 1, l_max + 1] matrix

    USAGE EXAMPLE:
        sc_filter = filter_sh_gaussian(cs, 500)

    HISTORY:
        - 18/12/2015 Feng Wei, State Key Laboratory of Geodesy and Earth's Dynamics &
            Institute of Geodesy and Geophysics, Chinese Academy of Sciences
        - 15/08/2024 Linus Shihora, GFZ, Potsdam
            porting to python
    '''

    
    #If input is list, convert to np.ndarray
    scs = np.asarray(scs)

    if scs.ndim not in [2, 3]:
        raise ValueError('Number of dimensions of gravity coeff. array must be 2 or 3.')
    
    if not isinstance(radius_filter, int):
        raise TypeError('Filter radius must be an integer.')
    
    if scs.ndim == 2:
        rows, cols = scs.shape
        
        if rows != cols:
            raise ValueError('Unexpected shape of gravity coeff. Expecting square array.')
        
        n_max = rows - 1 #Max. degree
        
        #Get weights of Gaussian
        weights = gaussian(n_max, radius_filter)

        #Transform to sc-format
        sc = cs_format(scs, 1)

        #Perform Gaussian smoothing
        sc_filter = np.zeros_like(sc)
        for l in range(n_max+1):
            sc_filter[l, :] = weights[l] * sc[l, :]

        #Transform back to cs-format
        cs_filter = cs_format(sc_filter, 1)


    else:   #scs.ndim == 3
        num_fields, rows, cols = scs.shape
        
        if rows != cols:
            raise ValueError('Unexpected shape of gravity coeff. Expecting square array.')
        
        n_max = rows - 1 #Max. degree
        
        #Get weights of Gaussian
        weights = gaussian(n_max, radius_filter)

        cs_filter = np.zeros_like(scs)
        for field in range(num_fields):
            #Transform to sc-format
            sc = cs_format(scs[field, :, :], 1)

            #Perform Gaussian smoothing
            sc_filter = np.zeros_like(sc)
            for l in range(n_max+1):
                sc_filter[l, :] = weights[l] * sc[l, :]

            #Transform back to cs-format
            cs_filter[field, :, :] = cs_format(sc_filter, 1)
    
    return cs_filter







def filter_sh_swenson(sc: np.ndarray | list) -> np.ndarray:
    '''
    FILTER_SH_SWENSON: Apply Swenson smoothing to spherical harmonic coefficients. Based on work in
                        'Swenson, 2006, Post-processing removal of correlated errors in GRACE data' and
                        'Duan, 2009, On the postprocessing removal of correlated errors in GRACE temporal
                        gravity field solutions'.

    IN:
        - sc            : Gravity coefficients in sc-format, i.e. a [l_max+1, 2l_max+1] matrix, cs-format
                          also suppoted

    OUT:
        - scs_filter    : Gravity coefficients in sc-format /S|C\\, 
                          i.e. a [l_max + 1, 2l_max + 1] matrix

    USAGE EXAMPLE:
        sc_filter = filter_sh_swenson(sc)

    HISTORY:
        - 09/05/2015 Feng Wei, State Key Laboratory of Geodesy and Earth's Dynamics &
            Institute of Geodesy and Geophysics, Chinese Academy of Sciences
        - 20/08/2024 Linus Shihora, GFZ, Potsdam
            porting to python
    '''

    #Check input format
    rows, cols = sc.shape
    if rows == cols:    #Input is in cs-format and needs to be converted to sc
        sc = cs_format(sc, 1)
    elif cols - (2*rows) == -1: #Input in sc-format already
        pass
    else:
        raise ValueError(f'Unsipported input shape of {rows}x{cols}.')
    

    #Determine maximum degree
    l_max = rows - 1

    if l_max == 60:     #Applied to old RL05 CSR GSM files
        deg_start = 5
        deg_end = 60
        ord_start = 5
        ord_end = 52
    elif l_max == 96:   #Applied to newer GSM files
        deg_start = 5
        deg_end = 96
        ord_start = 5
        ord_end = 88
    elif l_max == 90:   #Applied to some JPL GSM files
        deg_start = 5
        deg_end = 90
        ord_start = 5
        ord_end = 82
    else:
        raise ValueError('Only supports degrees of 60, 90 or 96.')
        

    #Prepare arrays
    sc_s = np.zeros((l_max+1, 2*l_max+1))   #To keep the fitted coefficients which will be removed in the end

    # sc_s[ord_end:deg_end+1, :] = sc[ord_end:deg_end+1, :]   #Unchanged part of the SHs
    sc_s[:deg_start, :] = sc[:deg_start, :]   #Unchanged part of the SHs


    #Calculate for each order separately
    for mm in range(ord_start, ord_end + 1):

        clm_col_s = np.zeros(l_max + 1)
        slm_col_s = np.zeros(l_max + 1)

        n_smooth = round(30 * np.exp(-mm / 10) + 1)  #Filter window, refer to Duan et al. (2009) eq. 1
        n_smooth = max(n_smooth, 5)

        if l_max == 96 and mm == 88:
            llvec = np.array([88,90,92,94,96])
        elif l_max == 60 and mm == 52:
            llvec = np.array([52,54,56,58,60])
        elif l_max == 90 and mm == 82:
            llvec = np.array([82,84,86,88,90])
        else:
            if mm > deg_start:
                llvec = np.arange(mm, deg_end + 1)
            else:
                llvec = np.arange(deg_start, deg_end + 1)

        for ll in llvec:
            #Constructing the n-vector. If ll is even (odd), n will contain even (odd) degrees.
            if ll - (n_smooth - 1) <= mm:
                if ll % 2 == 0: #ll is even
                    if mm % 2 == 0: #mm is even
                        n = np.arange(mm, mm + (n_smooth - 1) * 2 + 1, 2)
                    else:           #mm is odd
                        n = np.arange(mm + 1, mm + 1 + (n_smooth - 1) * 2 + 1, 2)
                else:   #ll is odd
                    if mm % 2 == 0: #mm is even
                        n = np.arange(mm + 1, mm + 1 + (n_smooth - 1) * 2 + 1, 2)
                    else:           #mm is odd
                        n = np.arange(mm, mm + (n_smooth - 1) * 2 + 1, 2)
            
            else:
                if n_smooth % 2 == 0:   #n_smooth is even
                    if ll + n_smooth <= deg_end:
                        n = np.arange(ll - n_smooth + 2, ll + n_smooth + 1, 2)
                    else:
                        if ll % 2 == 0: #even degrees
                            n = np.arange(deg_end - (n_smooth - 1) * 2, deg_end + 1, 2)
                        else:   #odd degrees
                            n = np.arange(deg_end - 1 - (n_smooth - 1) * 2, deg_end, 2)
                else:   #n_smooth is odd
                    if ll + n_smooth - 1 <= deg_end:
                        n = np.arange(ll - n_smooth + 1, ll + n_smooth, 2)
                    else:
                        if ll % 2 == 0:
                            n = np.arange(deg_end - (n_smooth - 1) * 2, deg_end + 1, 2)
                        else:
                            n = np.arange(deg_end - 1 - (n_smooth - 1) * 2, deg_end, 2)

            #Fill a vector for order mm with the coefficients for odd or even degrees within the window
            clm = sc[n, l_max + mm]
            slm = sc[n, l_max - mm]

            #Fit a second order polynomial
            poly_ord = 2
            poly_clm = np.polynomial.Polynomial.fit(n, clm, poly_ord)
            poly_slm = np.polynomial.Polynomial.fit(n, slm, poly_ord)

            #Get fit-results
            clm_col_s[ll] = poly_clm(ll)
            slm_col_s[ll] = poly_slm(ll)

        sc_s[:, l_max + mm] = clm_col_s
        sc_s[:, l_max - mm] = slm_col_s

    for mm in range(0, ord_start):

        clm_col_s = np.zeros(l_max + 1)
        slm_col_s = np.zeros(l_max + 1)

        n_smooth = round(30 * np.exp(-mm / 10) + 1)  #Filter window, refer to Duan et al. (2009)
        n_smooth = max(n_smooth, 5)

        llvec = np.arange(deg_start, deg_end + 1)

        for ll in llvec:
            if ll - (n_smooth-1) <= deg_start:
                if ll % 2 == 0:
                    if deg_start % 2 == 0:
                        n = np.arange(deg_start, deg_start + (n_smooth - 1) * 2 + 1, 2)
                    else:
                        n = np.arange(deg_start + 1, deg_start + 1 + (n_smooth - 1) * 2 + 1, 2)
                else:
                    if deg_start % 2 == 0:
                        n = np.arange(deg_start + 1, deg_start + 1 + (n_smooth - 1) * 2 + 1, 2)
                    else:
                        n = np.arange(deg_start, deg_start + (n_smooth - 1) * 2 + 1, 2)
            else:
                if n_smooth % 2 == 0:
                    if ll + n_smooth <= deg_end:
                        n = np.arange(ll - n_smooth + 2, ll + n_smooth + 1, 2)
                    else:
                        if ll % 2 == 0:
                            n = np.arange(deg_end - (n_smooth - 1) * 2, deg_end + 1, 2)
                        else:
                            np.arange(deg_end - 1 - (n_smooth - 1) * 2, deg_end, 2)
                else:
                    if ll + n_smooth - 1 <= deg_end:
                        n = np.arange(ll - n_smooth + 1, ll + n_smooth, 2)
                    else:
                        if ll % 2 == 0:
                            n = np.arange(deg_end - (n_smooth - 1) * 2, deg_end + 1, 2)
                        else:
                            n = np.arange(deg_end - 1 - (n_smooth - 1) * 2, deg_end, 2)

            n = n[n <= l_max]
            n = n[n >= 0]

            #Fill a vector for order mm with the coefficients for odd or even degrees within the window
            clm = sc[n, l_max + mm]
            slm = sc[n, l_max - mm]

            #Fit a second order polynomial
            poly_ord = 2
            poly_clm = np.polynomial.Polynomial.fit(n, clm, poly_ord)
            poly_slm = np.polynomial.Polynomial.fit(n, slm, poly_ord)

            #Get fit-results
            clm_col_s[ll] = poly_clm(ll)
            slm_col_s[ll] = poly_slm(ll)

        sc_s[:, l_max + mm] = clm_col_s
        sc_s[:, l_max - mm] = slm_col_s

    #Return input field minus fitted coefficients
    return sc - sc_s




def triplot(scs: np.ndarray | list, nmax: int) -> list:
    sc = cs_format(scs[:nmax + 1, :nmax + 1], 1)
    sc = np.flipud(sc)

    fig, ax = plt.subplots()
    c = ax.imshow(np.log10(np.abs(sc)), extent=[-nmax, nmax, 0, nmax], cmap='viridis')
    c.set_clim(np.log10(np.abs(sc[sc != 0]).min()), np.log10(np.abs(sc).max()))
    plt.xlabel('SH order', fontsize=15)
    plt.ylabel('SH degree', fontsize=15)
    h = plt.colorbar(c, ax=ax)
    h.set_label('log10', fontsize=15)
    plt.axis('equal')
    plt.axis('tight')
    plt.gca().tick_params(axis='both', which='major', labelsize=12)
    ax.invert_yaxis()

    return fig, ax, c






def cs2ddkformat(cs: np.ndarray) -> dict[np.ndarray, np.ndarray]:
    '''
    CS2DDKFORMAT: Transform spherical harmonic coefficients in cs-format to DDK-format

    IN:
        - cs              : Gravity coefficients in cs-format, i.e. a [l_max+1, l_max+1] matrix

    OUT:
        - sh_ddkformat    : C_lm and S_lm coefficients in DDK-format as a dict

    USAGE EXAMPLE:
        sh_ddk = cs2ddkformat(cs)

    HISTORY:
        - 09/07/2017 Feng Wei, State Key Laboratory of Geodesy and Earth's Dynamics &
            Institute of Geodesy and Geophysics, Chinese Academy of Sciences
        - 08/2024 Simon Schiller, TU Munich
            porting to python
        - 12/2024 Linus Shihora, GFZ, Potsdam
            stability and formatting
    '''

    # Check input format in terms of rows and columns
    if cs.ndim == 3:
        ntime, rows, cols = cs.shape
    elif cs.ndim == 2:
        rows, cols = cs.shape
        ntime = 1
    else:
        raise ValueError(f'Unsipported input shape of {cs.shape}.')

    if cols != rows:
        raise ValueError(f'Unsipported input shape of {rows}x{cols}.')


    l_max = rows - 1



    # Initialization of output:
    sh_ddkformat = {
        'C': np.zeros((l_max + 1, l_max + 1, ntime)),
        'S': np.zeros((l_max + 1, l_max + 1, ntime))
    }

    # Transform from cs to sc-format
    for ii in range(ntime):
        
        if cs.ndim == 3:
            cs_tmp = cs[ii, :, :]
        elif cs.ndim == 2:
            cs_tmp = cs

        c = np.tril(cs_tmp)
        s = np.rot90(np.triu(cs_tmp, 1), -1)
        sc_tmp = np.hstack([s[:, 1:l_max + 1], c])


    # Transform to DDK-format
        sh_ddkformat['C'][:, :, ii] = sc_tmp[:, l_max:2 * l_max + 1]
        sh_ddkformat['S'][1:l_max + 1, 1:l_max + 1, ii] = np.fliplr(sc_tmp[1:l_max + 1, 0:l_max])

    return sh_ddkformat








def ddk2csformat(sh_ddkformat: dict[np.ndarray, np.ndarray]) -> np.ndarray:
    '''
    DDK2CSFORMAT: Transform spherical harmonic coefficients from DDK-format to cs-format

    IN:
        - sh_ddkformat    : C_lm and S_lm coefficients in DDK-format as a dict

    OUT:
        - cs              : Gravity coefficients in cs-format, i.e. a [l_max+1, l_max+1] matrix

    USAGE EXAMPLE:
        cs = ddk2csformat(sh_ddk)

    HISTORY:
        - 09/07/2017 Feng Wei, State Key Laboratory of Geodesy and Earth's Dynamics &
            Institute of Geodesy and Geophysics, Chinese Academy of Sciences
        - 08/2024 Simon Schiller, TU Munich
            porting to python
        - 12/2024 Linus Shihora, GFZ, Potsdam
            stability and formatting
    '''

    # Initialize required variables l_max, ntime and cs-array
    l_max = sh_ddkformat['C'].shape[0] - 1

    if len(sh_ddkformat['C'].shape) == 3:
        ntime = sh_ddkformat['C'].shape[2]
    else:
        ntime = 1

    cs = np.zeros((ntime, l_max + 1, l_max + 1))



    # C and S coefficients from the dict into the combined cs-format:
    for ii in range(ntime):
        sc_tmp = np.zeros((l_max + 1, 2 * l_max + 1))
        sc_tmp[:, l_max:2 * l_max+1] = sh_ddkformat['C'][:, :, ii] if ntime > 1 else sh_ddkformat['C']
        
        #TODO: this is not easy to read. Needs to be refactored..
        sc_tmp[1:l_max + 1, :l_max] = np.fliplr(sh_ddkformat['S'][1:l_max + 1, 1:l_max + 1, ii]) if ntime > 1 else np.fliplr(sh_ddkformat['S'][1:l_max+1, 1:l_max+1])

        cs_tmp = cs_format(sc_tmp, 1)
        cs[ii, :, :] = cs_tmp

    # If there is only one epoch, remove superfluous time-axis
    cs = np.squeeze(cs)

    return cs





def read_BIN(ifile: str, *args) -> dict[np.ndarray, np.ndarray, np.ndarray]:
    '''
    READ_BIN: Read in a binary file containing symmetric/full or block diagonal matrices and 
                associated vectors and parameters. 

    IN:
        - ifile : Input file name
        - *args : dat=read_BIN(file) returns a structure array 'dat' with the file content 
            the matrix remains in packed form (dat.pack1 field) or: dat=read_BIN(file,'F') 
            also expands the matrix to its full form (dat.mat1 field).

    OUT:
        - dat   : Data from the binary file

    USAGE EXAMPLE:
        dat = read_BIN('Wbd_2-120.a_1d12p_4')

    HISTORY:
        - 2008 Roelof Rietbroek
        - 08/2024 Simon Schiller, TU Munich
            porting to python
        - 12/2024 Linus Shihora, GFZ, Potsdam
            stability and formatting
    '''


    # Opening and importing data
    # Open file for read access in little endian format:
    unpack = False

    for arg in args:
        if arg == 'F':
            unpack = True
    try:
        fid = open(ifile, 'rb')
    except IOError as e:
        print(e)
        return None


    # Read binary version and type from file:
    dat = {}
    dat['version'] = np.fromfile(fid, dtype='uint8', count=8).tobytes().decode('utf-8')
    dat['type'] = np.fromfile(fid, dtype='uint8', count=8).tobytes().decode('utf-8')
    dat['descr'] = np.fromfile(fid, dtype='uint8', count=80).tobytes().decode('utf-8')


    # Process indices
    # Read indices --> integers:inint,indbls,inval1,inval2,ipval1,ipval2
    metaint = np.fromfile(fid, dtype='int32', count=6)


    # Put index data in structure array:
    dat['nval1'], dat['nval2'], dat['pval1'], dat['pval2'] = metaint[2:]


    # Type dependent index data:
    if dat['type'] in {'BDSYMV0_', 'BDFULLV0'}:
        dat['nblocks'] = int(np.fromfile(fid, dtype='int32', count=1))


    # Process metadata
    # Integers:
    n_ints = int(metaint[0])
    n_dbls = int(metaint[1])

    if n_ints > 0:
        list_ints = np.fromfile(fid, dtype='uint8', count=n_ints * 24).reshape((n_ints, 24))
        dat['ints_d'] = list_ints
        dat['ints'] = np.fromfile(fid, dtype='int32', count=n_ints)


    # Doubles:
    if n_dbls > 0:
        list_dbls = np.fromfile(fid, dtype='uint8', count=n_dbls * 24).reshape((n_dbls, 24))
        dat['dbls_d'] = list_dbls
        dat['dbls'] = np.fromfile(fid, dtype='float64', count=n_dbls)


    # Side description meta data:
    list_side1 = np.fromfile(fid, dtype='uint8', count=dat['nval1'] * 24).reshape((dat['nval1'], 24))
    dat['side1_d'] = list_side1


    # Type specific meta data: 'BDSYMV0_', 'BDFULLV0'
    if dat['type'] in {'BDSYMV0_', 'BDFULLV0'}:
        dat['blockind'] = np.fromfile(fid,
                                      dtype='int32', count=dat['nblocks'])

    # Data (type dependent): SYMV0___
    if dat['type'] == 'SYMV0___':
        dat['pack1'] = np.fromfile(fid, dtype='float64', count=dat['pval1'])

        if unpack:
            row = np.arange(1, dat['nval1'] + 1)
            col = row
            ind = np.triu_indices(dat['nval1'])
            dat['mat1'] = np.zeros((dat['nval1'], dat['nval1']))
            dat['mat1'][ind] = dat['pack1']
            dat['mat1'] += dat['mat1'].T - np.diag(np.diag(dat['mat1']))
            del dat['pack1']


    # Data (type dependent): SYMV1___
    elif dat['type'] == 'SYMV1___':
        dat['vec1'] = np.fromfile(fid, dtype='float64', count=dat['nval1'])
        dat['pack1'] = np.fromfile(fid, dtype='float64', count=dat['pval1'])

        if unpack:
            row = np.arange(1, dat['nval1'] + 1)
            col = row
            ind = np.triu_indices(dat['nval1'])
            dat['mat1'] = np.zeros((dat['nval1'], dat['nval1']))
            dat['mat1'][ind] = dat['pack1']
            dat['mat1'] += dat['mat1'].T - np.diag(np.diag(dat['mat1']))
            del dat['pack1']


    # Data (type dependent): SYMV2___
    elif dat['type'] == 'SYMV2___':
        dat['vec1'] = np.fromfile(fid, dtype='float64', count=dat['nval1'])
        dat['vec2'] = np.fromfile(fid, dtype='float64', count=dat['nval1'])
        dat['pack1'] = np.fromfile(fid, dtype='float64', count=dat['pval1'])

        if unpack:
            row = np.arange(1, dat['nval1'] + 1)
            col = row
            ind = np.triu_indices(dat['nval1'])
            dat['mat1'] = np.zeros((dat['nval1'], dat['nval1']))
            dat['mat1'][ind] = dat['pack1']
            dat['mat1'] += dat['mat1'].T - np.diag(np.diag(dat['mat1']))
            del dat['pack1']


    # Data (type dependent): BDSYMV0_
    elif dat['type'] == 'BDSYMV0_':
        dat['pack1'] = np.fromfile(fid, dtype='float64', count=dat['pval1'])

        if unpack:
            dat['mat1'] = np.array([]).reshape(0, 0)
            skip = 0
            skipentries = 0

            for i in range(dat['nblocks']):
                sz = dat['blockind'][i] - skip
                row = np.arange(1, sz + 1)
                col = row
                ind = np.triu_indices(sz)
                ind = np.triu(ind[0] + skipentries, ind[1] + skipentries)
                skip = dat['blockind'][i]
                blk = np.zeros((sz, sz))
                blk[ind] = dat['pack1'][skipentries:skipentries + int(sz * (sz + 1) / 2)]
                blk += blk.T - np.diag(np.diag(blk))
                dat['mat1'] = np.block([[dat['mat1'], np.zeros((dat['mat1'].shape[0], blk.shape[1]))],
                                        [np.zeros((blk.shape[0], dat['mat1'].shape[1])), blk]])
                skipentries += int(sz * (sz + 1) / 2)

            del dat['pack1']


    # Data (type dependent): BDFULLV0
    elif dat['type'] == 'BDFULLV0':
        dat['pack1'] = np.fromfile(fid, dtype='float64', count=dat['pval1'])

        if unpack:
            dat['mat1'] = np.array([]).reshape(0, 0)
            skip = 0
            skipentries = 0

            for i in range(dat['nblocks']):
                sz = dat['blockind'][i] - skip
                blk = np.reshape(dat['pack1'][skipentries:skipentries + sz ** 2], (sz, sz))
                dat['mat1'] = np.block([[dat['mat1'], np.zeros((dat['mat1'].shape[0], blk.shape[1]))],
                                        [np.zeros((blk.shape[0], dat['mat1'].shape[1])), blk]])
                skip = dat['blockind'][i]
                skipentries += sz ** 2

            del dat['pack1']


    # Data (type dependent): FULLSQV0
    elif dat['type'] == 'FULLSQV0':
        dat['pack1'] = np.fromfile(fid, dtype='float64', count=dat['pval1'] ** 2)

        if unpack:
            dat['mat1'] = np.reshape(dat['pack1'], (dat['nval1'], dat['nval1']))
            del dat['pack1']


    fid.close()
    return dat








def filter_sh_ddk(number: int, data: np.ndarray) -> np.ndarray:
    '''
    FILTER_SH_DDK: Apply a DDK filter to input spherical harmonic coefficients

    IN:
        - number    : the type of DDK filter
            - DDK1d11: filtered with inverse signal degree power law 1e11*deg^4 (DDK5) weakest smoothing
            - DDK5d11: filtered with inverse signal degree power law 5e11*deg^4 (DDK4)
            - DDK1d12: filtered with inverse signal degree power law 1e12*deg^4 (DDK3)
            - DDK1d13: filtered with inverse signal degree power law 1e13*deg^4 (DDK2)
            - DDK1d14: filtered with inverse signal degree power law 1e14*deg^4 (DDK1) strongest smoothing 
        - data      : Gravity coefficients in cs-format, i.e. a [l_max+1, l_max+1] matrix

    OUT:
        - dataDDK   : Filtered gravity coefficients in cs-format, i.e. a [l_max+1, l_max+1] matrix

    USAGE EXAMPLE:
        cs_ddk = filter_sh_ddk(1, cs)

    HISTORY:
        - 2016 Roelof Rietbroek, This function is created based on the code from Roelof Rietbroek.
            Original code is found under: https://github.com/strawpants/GRACE-filter
            License found under: https://github.com/strawpants/GRACE-filter/blob/master/LICENSE
            Based on theoretical work described in Kusche et al. (2009, 10.1007/s00190-009-0308-3)
        - 18/10/2016 Feng Wei, State Key Laboratory of Geodesy and Earth's Dynamics &
            Institute of Geodesy and Geophysics, Chinese Academy of Sciences
        - 08/2024 Simon Schiller, TU Munich
            porting to python
        - 12/2024 Linus Shihora, GFZ, Potsdam
            stability and formatting
    '''

    # DDK filters require input binaries. One for each filter.
    # Selecting of the DDK binary based on the filter number (1 = strongest --> 8 = weakest)
    if number == 1:
        file = './data/Wbd_2-120.a_1d14p_4'
    elif number == 2:
        file = './data/Wbd_2-120.a_1d13p_4'
    elif number == 3:
        file = './data/Wbd_2-120.a_1d12p_4'
    elif number == 4:
        file = './data/Wbd_2-120.a_5d11p_4'
    elif number == 5:
        file = './data/Wbd_2-120.a_1d11p_4'
    elif number == 6:
        file = './data/Wbd_2-120.a_5d10p_4'
    elif number == 7:
        file = './data/Wbd_2-120.a_1d10p_4'
    elif number == 8:
        file = './data/Wbd_2-120.a_5d9p_4'
    else:
        raise ValueError(f'Invalid filter type {number}.')

    # Read binary file:
    dat = read_BIN(file)


    # Block interpretation of input data
    sz = [dat['blockind'][0]]
    nstart = [1]
    nend = [1 + sz[0] ** 2 - 1]

    for ij in range(1, dat['nblocks']):
        sz.append(dat['blockind'][ij] - dat['blockind'][ij - 1])
        nstart.append(1 + sum(s ** 2 for s in sz[:-1]))
        nend.append(nstart[-1] + sz[-1] ** 2 - 1)


    # Initialization of the maximum d/o and temporal dimension of the GRACE 
    # dataset from the spherical harmonic coefficients provided:
    nmax = data['C'].shape[0] - 1
    ntime = data['C'].shape[2]


    # Apply the DDK filter
    # Initialization of the Filtered equi-angular grid:
    dataDDK = data.copy()


    # Treatment of the coefficients for case m = 0
    ordre = 0
    ij = ordre + 1
    ss = nstart[ij - 1]
    ee = nend[ij - 1]
    block = np.reshape(dat['pack1'][nstart[ij - 1]-1:nend[ij - 1]], (sz[ij], sz[ij])).T
    block = block[:nmax - 1, :nmax - 1]

    for klm in range(ntime):            # If several points in time are available
        coef = data['C'][2:, ordre, klm]
        dataDDK['C'][2:, ij - 1, klm] = np.dot(block, coef)


    # Treatment of the coefficients for case m != 0
    ordre = 1
    while ordre < nmax + 1:
        ij = 2 * ordre
        ss = nstart[ij]
        ee = nend[ij]
        szsz = sz[ij+1];
        blockC = np.reshape(dat['pack1'][nstart[ij - 1]-1:nend[ij - 1]], (sz[ij-1], sz[ij-1])).T
        blockS = np.reshape(dat['pack1'][nstart[ij]-1:nend[ij]], (sz[ij], sz[ij])).T

        fin = min(nmax - 1, nmax - ordre + 1)
        blockC = blockC[:fin, :fin]
        blockS = blockS[:fin, :fin]

        deb = max(3, ordre + 1)
        for klm in range(ntime):        # If several points in time are available
            coefC = data['C'][deb - 1:, ordre, klm]
            coefS = data['S'][deb - 1:, ordre, klm]
            dataDDK['C'][deb - 1:, ordre, klm] = np.dot(blockC, coefC)
            dataDDK['S'][deb - 1:, ordre, klm] = np.dot(blockS, coefS)

        ordre += 1

    return dataDDK